use serial;
use serial::prelude::*;
pub use serial::PortSettings;

use utilities::prelude::*;

use std::cell::RefCell;
use std::ffi::OsStr;
use std::time::Duration;

pub struct Port {
    serial_port: RefCell<Box<dyn SerialPort>>,
}

impl Port {
    pub fn open<T: AsRef<OsStr> + ?Sized + std::fmt::Display>(
        port: &T,
        settings: PortSettings,
    ) -> VerboseResult<Self> {
        let mut port = serial::open(port).map_err(|err| {
            format!(
                "could not open serial port ({:?}, {}, {})",
                err.kind(),
                err,
                port
            )
        })?;

        port.configure(&settings)
            .map_err(|err| format!("failed configuring serial port ({})", err))?;

        port.set_timeout(Duration::from_millis(2500))
            .map_err(|err| format!("failed setting time out for serial port ({})", err))?;

        Ok(Port {
            serial_port: RefCell::new(Box::new(port)),
        })
    }

    pub fn read(&self) -> VerboseResult<String> {
        let mut buf: Vec<u8> = (0..255).collect();

        self.serial_port
            .try_borrow_mut()?
            .read(&mut buf[..])
            .map_err(|err| format!("failed reading serial port ({})", err))?;

        Ok(String::from_utf8(buf)
            .map_err(|err| format!("failed converting utf8 buffer ({})", err))?)
    }

    pub fn write(&self, msg: &str) -> VerboseResult<()> {
        self.serial_port
            .try_borrow_mut()?
            .write(msg.as_bytes())
            .map_err(|err| format!("failed writing to serial port ({})", err))?;

        Ok(())
    }
}
